#ifndef __BITVECTOR_H
#define __BITVECTOR_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stdbool.h>

#include <malloc.h>

typedef unsigned int chunk_t;
typedef unsigned int index_t;

/* You can use bitvector_set_allocator below to set new memory management
 * functions, which should match the below types. The reason that free has an
 * extra size parameter is to make it easier to use with mmap and munmap. */
typedef void* (*malloc_ptr_t)(size_t);
typedef void* (*realloc_ptr_t)(void*, size_t);
typedef void (*free_ptr_t)(void*, size_t);

/* It is better to not change this structure. It is not guarenteed to be
 * stable. Use the provided functions and macros. */
struct bitvector_s
{
	/* Pointer to an array of uints: the actual bits are stored here */
	chunk_t *bits;

	/* The number of bits in the array. This is always rounded up to chunk
	 * boundries */
	index_t size;

	/* The number of elements in the array */
	index_t arraylen;

	/* pointer to the allocater used to allocate only the bit data for this
	 * bitvector. This defaults to the system malloc */
	malloc_ptr_t malloc;
	/* Pointer to the realloc function to be used to resize this object if
	 * necessary */
	realloc_ptr_t realloc;
	/* Pointer to the function used to free the blocks allocated from either of
	 * the above */
	free_ptr_t free;

	struct bitvector_cache_s
	{
		/* A cached index: the first bit set. This is valid if dirty is false */
		index_t firstset;
		/* This is true if none is set */
		bool firstset_none;
		/* A cached index: the first bit unset. This is valid if dirty is false */
		index_t firstunset;
		/* This is true if none is unset */
		bool firstunset_none;
	} cache;

	/* True if the variables may have been changed and need to be recalculated
	 * */
	bool dirty;
};

typedef struct bitvector_s bitvector;

/* The number of bytes of each individual element in the array that stores the
 * bits */
#define BV_CHUNKSIZE_BYTES		(sizeof(chunk_t))
#define BV_CHUNKSIZE			(BV_CHUNKSIZE_BYTES<<3)

/* The number of bits availible for use in the bitvector a */
#define BV_BITSIZE(a)			(a->size)
/* The number of bytes availible for use in the bitvector a */
#define BV_BYTESIZE(a)			(a->size>>3)

/* Retrieve a pointer to the array of bits */
#define BV_BITS(a)				(a->bits)

/* Return a pointer to a new bitvector of size s, or return NULL on failure */
bitvector *bitvector_create (index_t s);
/* Return a pointer to a new bitvector of size s, or return NULL on failure.
 * The memory for the bits stored will be managed by the provided function
 * pointers. This may be used for mapping to a file, or something. The
 * bitvector is guaranteed to use a single contiguous memory block. */
bitvector *bitvector_create_allocator (index_t s, malloc_ptr_t, realloc_ptr_t,
		free_ptr_t);
/* Same as above, but do not initialize the bitvector data to all zeros */
bitvector *bitvector_create_dirty (index_t size);
/* Frees the bitvector b */
void bitvector_free (bitvector *b);

/* Set the functions to be used to manage memory for the bits in the bitvector */
void bitvector_set_default_allocator (malloc_ptr_t, realloc_ptr_t, free_ptr_t);
/* Reset the allocator to the default set */
void bitvector_reset_default_allocator ();

/* Sets bit #n to true, in the vector b */
void bitvector_set (bitvector *b, index_t n);
/* Sets bit #n to false, in the vector b */
void bitvector_unset (bitvector *b, index_t n);
/* Returns 0 if bit #n is unset in the vector b, and nonzero otherwise */
int bitvector_get (bitvector *b, index_t n);

/* Stores a string ("10110011"-style) representation of the bitvector b, and
 * stores it in the provided buffer. It is the users responsibility to provide
 * a buffer of sufficient size, which is: BV_BITSIZE(a)+1. The +1 is because it
 * is null-terminated. */
void bitvector_tostring (bitvector *b, char *buffer);
/* The above, but in reverse */
bitvector *bitvector_fromstring (char *s);
/* Read all the bits of an integer into a new bitvector */
bitvector *bitvector_fromint (chunk_t n);

/* Convert it to a string which is has no NULLs in it, and therefore is a valid
 * C-string (good for storing in databases, etc */
char *bitvector_tocstring (bitvector *b);
bitvector *bitvector_fromcstring (char *s);

/* Resize the bitvector b to be able to store n bytes. The requested size is
 * rounded up to the nearest chunk size (see above, by the bitvector_s
 * definition, for a discussion of chunk). The bitvector is either zero-padded
 * or truncated to accomadate the new length. The _ns function does not copy
 * the old data. No initialization is performed on it.
 *
 * Zero (0) is returned on success, and non-zero on failure. */

int bitvector_resize (bitvector *b, index_t n);
int bitvector_resize_ns (bitvector *b, index_t n);

/* In _copy, the user supplies a bitvector struct, already allocated. In
 * _duplicate(), a new one is returned, and it is the users responsibility to
 * bitvector_free() it. */
int bitvector_copy (bitvector *src, bitvector *dst);
bitvector *bitvector_duplicate (bitvector *b);

/* dest = lhs & rhs */
int bitvector_and (bitvector *dest, bitvector *lhs, bitvector *rhs);
/* lhs &= rhs */
void bitvector_andeq (bitvector *lhs, bitvector *rhs);
/* dest = lhs | rhs */
int bitvector_or (bitvector *dest, bitvector *lhs, bitvector *rhs);
/* lhs |= rhs */
int bitvector_oreq (bitvector *lhs, bitvector *rhs);
/* dest = lhs ^ rhs */
int bitvector_xor (bitvector *dest, bitvector *lhs, bitvector *rhs);
/* lhs ^= rhs */
int bitvector_xoreq (bitvector *lhs, bitvector *rhs);
/* nb = ~b. This is more efficient than _copy() and _invert() */
int bitvector_not (bitvector *nb, bitvector *b);
/* b = ~b */
void bitvector_invert (bitvector *b);

/* Returns zero if b has any bets set to on or true. Returns nonzero otherwise
 * (Currently 1, but this may change.). */
int bitvector_isempty (bitvector *b);
int bitvector_isfull (bitvector *b);

/* Shift b left n bits. Left shifting is never destructive. Right shifted data
 * is lost if it passes beyond the decimal. */
void bitvector_leftshift (bitvector *b, index_t n);
void bitvector_rightshift (bitvector *b, index_t n);

/* Clears all bits. Does not resize */
void bitvector_clear (bitvector *b);

/* Equality check */
int bitvector_isequal (bitvector *a, bitvector *b);

/* Sets i to the first set bit, returning zero on success. Returns -1 on
 * failure */
int bitvector_firstset (bitvector *b, index_t *i);
/* Sets i to the first unset bit, returning zero on success. Returns -1 on
 * failure */
int bitvector_firstunset (bitvector *b, index_t *);

#ifdef __cplusplus
}
#endif

#endif
