#include <stdlib.h>
#include <stdio.h>

#if HAVE_LIBCHECK
#include <check.h>

#include "bitvector.h"

/*
Unit tests for the bitvector library. These tests were written on a system
where ints are 32-bits wide, and the tests take that into account. Don't do
this on a 64 bit system. Write new, kinder tests. This should be fixed
sometime.
   */

START_TEST(test_create)
{
	int i;
	char errmsg[200];

	for (i=0; i<100; i++)
	{
		bitvector *b;
		int j;

		b = bitvector_create (i);
		if (b == NULL)
		{
			strcpy (errmsg, "Bitvector creation failed: no bitvector*");
			break;
		} else if (BV_BITS(b) == NULL)
		{
			strcpy (errmsg, "Bitvector creation failed: no bits");
			break;
		} else if (BV_BITSIZE(b) <= i)
		{
			strcpy (errmsg, "Bitvector size not correct at creation");
			break;
		} else if (!b->cache.firstset_none)
		{
			strcpy (errmsg, "Bitvector firstset not correct at creation");
			break;
		} else if (b->cache.firstunset != 0)
		{
			strcpy (errmsg, "Bitvector firstunset not correct at creation");
			break;
		}

		for (j=0; j<b->arraylen; j++)
			if (BV_BITS(b)[j])
			{
				strcpy (errmsg, "Bitvector creation failed (Bitvector has bits set)");
				break;
			}

		bitvector_free (b);
	}
	if (i != 100)
		fail (errmsg);
}
END_TEST

START_TEST(test_clear)
{
	bitvector *b;

	b = bitvector_create (32);
	bitvector_set (b, 15);
	bitvector_set (b, 10);

	bitvector_clear (b);

	fail_unless (bitvector_get(b, 15)==0, "Bitvector clear failed (Bitvector bit #15 set)");
	fail_unless (bitvector_get(b, 10)==0, "Bitvector clear failed (Bitvector bit #10 set)");

	bitvector_free (b);
}
END_TEST

START_TEST(test_create_dirty)
{
	bitvector *b;
	int firstonbit, firstoffbit;

	b = bitvector_create_dirty (0);
	fail_unless (b != NULL, "Bitvector creation failed: no bitvector*");
	fail_unless (BV_BITS(b) != NULL, "Bitvector creation failed: no bits");
	fail_unless (BV_BITSIZE(b) == sizeof(chunk_t)*8, "Bitvector size not set correctly on creation");
	fail_unless (b->arraylen == 1, "Bitvector arraylen not set correctly on creation");

	bitvector_free (b);
}
END_TEST

START_TEST(test_set)
{
	bitvector *b;

	b = bitvector_create (64);
	bitvector_set (b, 4);
	fail_unless (BV_BITS(b)[0] & (1<<4), "Low bit setting failed");
	bitvector_set (b, 48);
	fail_unless (BV_BITS(b)[1] & (1<<(48-32)), "High bit setting failed");
	bitvector_free (b);
}
END_TEST

START_TEST(test_unset)
{
	bitvector *b;

	b = bitvector_create (64);
	bitvector_set (b, 4);
	bitvector_set (b, 8);
	bitvector_unset (b, 4);
	fail_unless (!bitvector_get(b, 4), "Low bit unsetting failed");
	bitvector_set (b, 48);
	bitvector_unset (b, 48);
	fail_unless (!bitvector_get(b, 48), "High bit unsetting failed");
	fail_unless (bitvector_get(b, 8), "Bit unsetting failed (Expected 8 set)");

	bitvector_set (b,5);
	bitvector_unset (b,5);
	bitvector_unset (b,5);
	fail_unless (!bitvector_get(b, 5), "Low bit unsetting failed (Expected 5 unset)");
	
	bitvector_free (b);
}
END_TEST

START_TEST(test_get)
{
	bitvector *b;

	b = bitvector_create (8);
	bitvector_set (b, 4);
	fail_unless (!bitvector_get (b, 0), "Bit retrieval failed (Expected false)");
	fail_unless (bitvector_get (b, 4), "Bit retrieval failed (Expected true)");
	bitvector_free (b);
}
END_TEST

START_TEST(test_tostring)
{
	bitvector *b;
	char result[64];
	int i;

	b = bitvector_create (31);

	for (i=0; i<BV_BITSIZE(b); i++)
		if (i%2)
			bitvector_set (b, i);
	bitvector_tostring (b, result);

	fail_unless (strcmp(result, "01010101010101010101010101010101") == 0,
			"Bitvector to string failed");

	bitvector_free (b);
}
END_TEST

START_TEST(test_fromstring)
{
	bitvector *b;
	char str[] = "00100100";

	b = bitvector_fromstring (str);

	fail_unless (bitvector_get (b, 2), "String to bitvector conversion failed");
	fail_unless (bitvector_get (b, 5), "String to bitvector conversion failed");

	bitvector_free (b);
}
END_TEST

START_TEST(test_fromint)
{
	bitvector *b;
	/* Set bits 0, 1, 3, 10 */
	chunk_t src = 1|2|8|1024;

	b = bitvector_fromint (src);

	fail_unless (bitvector_get (b, 0), "Int to bitvector conversion failed (Bit 0 unset)");
	fail_unless (bitvector_get (b, 1), "Int to bitvector conversion failed (Bit 1 unset)");
	fail_unless (bitvector_get (b, 3), "Int to bitvector conversion failed (Bit 3 unset)");
	fail_unless (bitvector_get (b, 10), "Int to bitvector conversion failed (Bit 10 unset)");

	bitvector_free(b);
}
END_TEST

START_TEST(test_resize)
{
	bitvector *b;

	b = bitvector_create (96);
	bitvector_set (b, 16);
	bitvector_set (b, 48);
	bitvector_resize (b, 64);
	fail_unless (BV_BITSIZE(b) == 64, "Bitvector shrink failed");
	fail_unless (bitvector_get (b, 16), "Bitvector shrink low data copy failed");
	fail_unless (bitvector_get (b, 48), "Bitvector shrink high data copy failed");

	bitvector_resize (b, 256);
	fail_unless (BV_BITSIZE(b) == 256, "Bitvector enlarge failed");
	fail_unless (bitvector_get (b, 16), "Bitvector enlarge low data copy failed");
	fail_unless (bitvector_get (b, 48), "Bitvector enlarge high data copy failed");
	bitvector_free (b);
}
END_TEST

START_TEST(test_resize_ns)
{
	bitvector *b;

	b = bitvector_create (96);
	bitvector_resize (b, 64);
	fail_unless (BV_BITSIZE(b) == 64, "Bitvector shrink failed");
	bitvector_resize (b, 256);
	fail_unless (BV_BITSIZE(b) == 256, "Bitvector enlarge failed");
	bitvector_free (b);
}
END_TEST

START_TEST(test_and)
{
	/* Test a = b & c */
	bitvector *a=bitvector_create (16);
	bitvector *b=bitvector_create (48);
	bitvector *c=bitvector_create (70);

	bitvector_set (b, 4);
	bitvector_set (b, 15);
	bitvector_set (c, 15);
	bitvector_set (b, 40);
	bitvector_set (c, 40);
	bitvector_set (c, 63);
	
	bitvector_and (a, b, c);

	fail_unless (bitvector_get(a, 4) == 0, "Bitvector 'and' operation failed (4 set)");
	fail_unless (bitvector_get(a, 63) == 0, "Bitvector 'and' operation failed (63 set)");
	fail_unless (bitvector_get(a, 15) != 0, "Bitvector 'and' operation failed (15 not set)");
	fail_unless (bitvector_get(a, 40) != 0, "Bitvector 'and' operation failed (40 not set)");

	bitvector_free (a);
	bitvector_free (b);
	bitvector_free (c);
}
END_TEST

START_TEST(test_andeq)
{
	/* Test b &= c */
	bitvector *b=bitvector_create (48);
	bitvector *c=bitvector_create (70);

	bitvector_set (b, 4);
	bitvector_set (b, 15);
	bitvector_set (c, 15);
	bitvector_set (b, 40);
	bitvector_set (c, 40);
	bitvector_set (c, 63);
	
	bitvector_andeq (b, c);

	fail_unless (bitvector_get(b, 4) == 0, "Bitvector 'and' operation failed (4 set)");
	fail_unless (bitvector_get(b, 63) == 0, "Bitvector 'and' operation failed (63 set)");
	fail_unless (bitvector_get(b, 15) != 0, "Bitvector 'and' operation failed (15 not set)");
	fail_unless (bitvector_get(b, 40) != 0, "Bitvector 'and' operation failed (40 not set)");

	bitvector_free (b);
	bitvector_free (c);
}
END_TEST

START_TEST(test_not)
{
	bitvector *b = bitvector_create (64);
	bitvector *nb = bitvector_create (0);

	bitvector_set (b, 8);
	bitvector_set (b, 16);
	bitvector_set (b, 32);
	bitvector_set (b, 48);
	bitvector_set (b, 63);

	bitvector_not (nb, b);

	fail_unless (bitvector_get(nb, 8) == 0, "Bitvector 'not' operation failed (8 set)");
	fail_unless (bitvector_get(nb, 16) == 0, "Bitvector 'not' operation failed (16 set)");
	fail_unless (bitvector_get(nb, 32) == 0, "Bitvector 'not' operation failed (32 set)");
	fail_unless (bitvector_get(nb, 48) == 0, "Bitvector 'not' operation failed (48 set)");
	fail_unless (bitvector_get(nb, 63) == 0, "Bitvector 'not' operation failed (63 set)");
	fail_unless (bitvector_get(nb, 3) != 0, "Bitvector 'not' operation failed (3 unset)");
	fail_unless (bitvector_get(nb, 55) != 0, "Bitvector 'not' operation failed (55 unset)");

	bitvector_free (b);
	bitvector_free (nb);
}
END_TEST

START_TEST(test_invert)
{
	bitvector *b = bitvector_create (64);

	bitvector_set (b, 8);
	bitvector_set (b, 16);
	bitvector_set (b, 32);
	bitvector_set (b, 48);
	bitvector_set (b, 63);

	bitvector_invert (b);

	fail_unless (bitvector_get(b, 8) == 0, "Bitvector 'invert' operation failed (8 set)");
	fail_unless (bitvector_get(b, 16) == 0, "Bitvector 'invert' operation failed (16 set)");
	fail_unless (bitvector_get(b, 32) == 0, "Bitvector 'invert' operation failed (32 set)");
	fail_unless (bitvector_get(b, 48) == 0, "Bitvector 'invert' operation failed (48 set)");
	fail_unless (bitvector_get(b, 63) == 0, "Bitvector 'invert' operation failed (63 set)");
	fail_unless (bitvector_get(b, 3) != 0, "Bitvector 'invert' operation failed (3 unset)");
	fail_unless (bitvector_get(b, 55) != 0, "Bitvector 'invert' operation failed (55 unset)");

	bitvector_free (b);
}
END_TEST

START_TEST(test_or)
{
	bitvector *a = bitvector_create (0);
	bitvector *b = bitvector_create (32);
	bitvector *c = bitvector_create (96);

	bitvector_set (b, 0);
	bitvector_set (c, 8);
	bitvector_set (b, 16);
	bitvector_set (c, 24);
	bitvector_set (b, 31);
	bitvector_set (c, 72);

	bitvector_or (a, b, c);

	/* Make sure that some randomly chosen bits aren't set, that shouldn't be */
	fail_unless (bitvector_get (a, 4) == 0, "Bitvector 'or' operation failed (4 set)");
	fail_unless (bitvector_get (a, 12) == 0, "Bitvector 'or' operation failed (12 set)");

	/* Make sure that all the bits that we set in either are set in the result */
	fail_unless (bitvector_get (a, 0) != 0, "Bitvector 'or' operation failed (0 not set)");
	fail_unless (bitvector_get (a, 8) != 0, "Bitvector 'or' operation failed (8 not set)");
	fail_unless (bitvector_get (a, 16) != 0, "Bitvector 'or' operation failed (16 not set)");
	fail_unless (bitvector_get (a, 24) != 0, "Bitvector 'or' operation failed (24 not set)");
	fail_unless (bitvector_get (a, 31) != 0, "Bitvector 'or' operation failed (31 not set)");
	fail_unless (bitvector_get (a, 72) != 0, "Bitvector 'or' operation failed (72 not set)");

	bitvector_free (a);
	bitvector_free (b);
	bitvector_free (c);
}
END_TEST

START_TEST(test_oreq)
{
	bitvector *b = bitvector_create (32);
	bitvector *c = bitvector_create (96);

	bitvector_set (b, 0);
	bitvector_set (c, 8);
	bitvector_set (b, 16);
	bitvector_set (c, 24);
	bitvector_set (b, 31);
	bitvector_set (c, 72);

	bitvector_oreq (b, c);

	/* Make sure that some randomly chosen bits aren't set, that shouldn't be */
	fail_unless (bitvector_get (b, 4) == 0, "Bitvector 'or-eq' operation failed (4 set)");
	fail_unless (bitvector_get (b, 12) == 0, "Bitvector 'or-eq' operation failed (12 set)");

	/* Make sure that all the bits that we set in either are set in the result */
	fail_unless (bitvector_get (b, 0) != 0, "Bitvector 'or-eq' operation failed (0 not set)");
	fail_unless (bitvector_get (b, 8) != 0, "Bitvector 'or-eq' operation failed (8 not set)");
	fail_unless (bitvector_get (b, 16) != 0, "Bitvector 'or-eq' operation failed (16 not set)");
	fail_unless (bitvector_get (b, 24) != 0, "Bitvector 'or-eq' operation failed (24 not set)");
	fail_unless (bitvector_get (b, 31) != 0, "Bitvector 'or-eq' operation failed (31 not set)");
	fail_unless (bitvector_get (b, 72) != 0, "Bitvector 'or-eq' operation failed (72 not set)");

	bitvector_free (b);
	bitvector_free (c);
}
END_TEST

START_TEST(test_xor)
{
	bitvector *a = bitvector_create (0);
	bitvector *b = bitvector_create (32);
	bitvector *c = bitvector_create (96);

	bitvector_set (b, 0);
	bitvector_set (c, 0);

	bitvector_set (b, 8);
	bitvector_set (c, 8);

	bitvector_set (b, 16);
	bitvector_set (b, 24);
	bitvector_set (c, 72);

	bitvector_xor (a, b, c);

	fail_unless (bitvector_get (a, 0) == 0, "Bitvector 'xor' operation failed (0 set)");
	fail_unless (bitvector_get (a, 8) == 0, "Bitvector 'xor' operation failed (8 set)");

	fail_unless (bitvector_get (a, 16) != 0, "Bitvector 'xor' operation failed (16 unset)");
	fail_unless (bitvector_get (a, 24) != 0, "Bitvector 'xor' operation failed (24 unset)");
	fail_unless (bitvector_get (a, 72) != 0, "Bitvector 'xor' operation failed (72 unset)");

	bitvector_free (a);
	bitvector_free (b);
	bitvector_free (c);
}
END_TEST

START_TEST(test_xoreq)
{
	bitvector *b = bitvector_create (32);
	bitvector *c = bitvector_create (96);

	bitvector_set (b, 0);
	bitvector_set (c, 0);

	bitvector_set (b, 8);
	bitvector_set (c, 8);

	bitvector_set (b, 16);
	bitvector_set (b, 24);
	bitvector_set (c, 72);

	bitvector_xoreq (b, c);

	fail_unless (bitvector_get (b, 0) == 0, "Bitvector 'xor-eq' operation failed (0 set)");
	fail_unless (bitvector_get (b, 8) == 0, "Bitvector 'xor-eq' operation failed (8 set)");

	fail_unless (bitvector_get (b, 16) != 0, "Bitvector 'xor-eq' operation failed (16 unset)");
	fail_unless (bitvector_get (b, 24) != 0, "Bitvector 'xor-eq' operation failed (24 unset)");
	fail_unless (bitvector_get (b, 72) != 1, "Bitvector 'xor-eq' operation failed (72 unset)");

	bitvector_free (b);
	bitvector_free (c);
}
END_TEST

START_TEST(test_copy)
{
	bitvector *src, *dst;
	src = bitvector_create (48);
	dst = bitvector_create (48);

	bitvector_set (src, 0);
	bitvector_set (src, 8);
	bitvector_set (src, 32);
	bitvector_set (src, 48);

	bitvector_copy (src, dst);

	fail_unless (BV_BITSIZE(dst) == BV_BITSIZE(src), "Bitvector copy failed (Sizes unequal)");
	fail_unless (dst->arraylen == src->arraylen, "Bitvector copy failed (Array lengths unequal)");

	fail_unless (bitvector_get (dst, 4) == 0, "Bitvector copy failed (4 set)");
	fail_unless (bitvector_get (dst, 16) == 0, "Bitvector copy failed (16 set)");

	fail_unless (bitvector_get (dst, 0) != 0, "Bitvector copy failed (0 not set)");
	fail_unless (bitvector_get (dst, 8) != 0, "Bitvector copy failed (8 not set)");
	fail_unless (bitvector_get (dst, 32) != 0, "Bitvector copy failed (32 not set)");
	fail_unless (bitvector_get (dst, 48) != 0, "Bitvector copy failed (48 not set)");

	bitvector_free (src);
	bitvector_free (dst);
}
END_TEST

START_TEST(test_duplicate)
{
	bitvector *src, *dst;
	src = bitvector_create (48);

	bitvector_set (src, 0);
	bitvector_set (src, 8);
	bitvector_set (src, 32);
	bitvector_set (src, 48);

	dst = bitvector_duplicate (src);

	fail_unless (BV_BITSIZE(dst) == BV_BITSIZE(src), "Bitvector duplicate failed (Sizes unequal)");
	fail_unless (dst->arraylen == src->arraylen, "Bitvector duplicate failed (Array lengths unequal)");

	fail_unless (bitvector_get (dst, 4) == 0, "Bitvector duplicate failed (4 set)");
	fail_unless (bitvector_get (dst, 16) == 0, "Bitvector duplicate failed (16 set)");

	fail_unless (bitvector_get (dst, 0) != 0, "Bitvector duplicate failed (0 not set)");
	fail_unless (bitvector_get (dst, 8) != 0, "Bitvector duplicate failed (8 not set)");
	fail_unless (bitvector_get (dst, 32) != 0, "Bitvector duplicate failed (32 not set)");
	fail_unless (bitvector_get (dst, 48) != 0, "Bitvector duplicate failed (48 not set)");

	bitvector_free (src);
	bitvector_free (dst);
}
END_TEST

START_TEST(test_isempty)
{
	bitvector *b;
	b = bitvector_create (16);
	fail_unless (bitvector_isempty(b) != 0, "Bitvector zero test failed (Expected zero)");
	bitvector_set (b, 4);
	fail_unless (bitvector_isempty(b) == 0, "Bitvector zero test failed (Expected nonzero)");
	bitvector_free (b);
}
END_TEST

START_TEST(test_isfull)
{
	bitvector *b = bitvector_create (16);
	int i;

	fail_unless (bitvector_isfull(b) == 0, "Bitvector full test failed (Expected zero)");

	for (i=0; i<BV_BITSIZE(b); i++)
		bitvector_set (b, i);

	fail_unless (bitvector_isfull(b) != 0, "Bitvector full test failed (Expected nonzero)");

	bitvector_free (b);
}
END_TEST

START_TEST(test_isequal)
{
	bitvector *a=bitvector_create (48);
	bitvector *b=bitvector_create (72);

	bitvector_set (a, 5);
	bitvector_set (a, 35);
	bitvector_set (b, 6);
	bitvector_set (b, 36);

	fail_unless (bitvector_isequal(a,b) == 0, "Bitvector quality test faileed (Expected zero)");

	bitvector_set (b, 5);
	bitvector_set (b, 35);
	bitvector_set (a, 6);
	bitvector_set (a, 36);

	fail_unless (bitvector_isequal(a,b) != 0, "Bitvector quality test faileed (Expected nonzero)");

	bitvector_free (a);
	bitvector_free (b);
}
END_TEST

START_TEST(test_leftshift)
{
	bitvector *b = bitvector_create (32);
	bitvector_set (b, 0);
	bitvector_set (b, 31);
	bitvector_leftshift (b, 1);

	fail_unless (bitvector_get (b, 0) == 0, "Bitvector left shift failed (Bit 0 set)");
	fail_unless (bitvector_get (b, 31) == 0, "Bitvector left shift failed (Bit 31 set)");
	fail_unless (bitvector_get (b, 1), "Bitvector left shift failed (Bit 1 unset)");
	fail_unless (bitvector_get (b, 32), "Bitvector left shift failed (Bit 32 unset)");

	bitvector_set (b, 31);
	bitvector_leftshift (b, 3);

	fail_unless (bitvector_get (b, 1) == 0, "Bitvector left shift failed (Bit 1 set)");
	fail_unless (bitvector_get (b, 31) == 0, "Bitvector left shift failed (Bit 31 set)");
	fail_unless (bitvector_get (b, 4), "Bitvector left shift failed (Bit 4 unset)");
	fail_unless (bitvector_get (b, 34), "Bitvector left shift failed (Bit 34 unset)");

	bitvector_leftshift (b, 0);

	fail_unless (bitvector_get (b, 34), "Bitvector left shift by zero failed (Bit 34 unset");
	fail_unless (BV_BITSIZE(b) == 64, "Bitvector left shift failed (Expected 64 bit size)");

	bitvector_clear (b);
	bitvector_set (b, 0);

	bitvector_leftshift (b, 40);
	
	fail_unless (bitvector_get (b, 40), "Bitvector left shift failed (Bit 40 unset)");

	bitvector_free (b);
}
END_TEST

START_TEST(test_rightshift)
{
	bitvector *b = bitvector_create (64);
	bitvector_set (b, 8);
	bitvector_set (b, 38);

	bitvector_rightshift (b, 7);

	fail_unless (bitvector_get (b, 8) == 0, "Bitvector right shift failed (Bit 8 set)");
	fail_unless (bitvector_get (b, 38) == 0, "Bitvector right shift failed (Bit 38 set)");
	fail_unless (bitvector_get (b, 1), "Bitvector right shift failed (Bit 1 unset)");
	fail_unless (bitvector_get (b, 31), "Bitvector right shift failed (Bit 31 unset)");

	bitvector_leftshift (b, 0);

	fail_unless (bitvector_get (b, 31), "Bitvector right shift by zero failed (Bit 31 unset");

	bitvector_clear (b);
	bitvector_set (b, 40);

	bitvector_rightshift (b, 40);
	
	fail_unless (bitvector_get (b, 0), "Bitvector right shift failed (Bit 0 unset)");
	bitvector_free (b);
}
END_TEST

START_TEST(test_cstring)
{
	char *s;
	bitvector *b = bitvector_create (64);
	bitvector *b2;

	bitvector_set (b, 10);
	bitvector_set (b, 11);
	bitvector_set (b, 12);
	bitvector_set (b, 30);
	bitvector_set (b, 40);
	bitvector_set (b, 63);

	s = bitvector_tocstring (b);
	b2 = bitvector_fromcstring (s);

	fail_unless (bitvector_isequal (b, b2), "Bitvector to C-String failed (input != output)");

	free (s);
	bitvector_free (b2);
	bitvector_free (b);
}
END_TEST

START_TEST(test_firstset)
{
	bitvector *a = bitvector_create(100);
	index_t idx;

	fail_unless(bitvector_firstset(a, &idx), "Bitvector first set failed when none set");

	bitvector_set(a, 45);
	fail_unless(bitvector_firstset(a, &idx) == 0 && idx == 45,
			"Bitvector first set failed when one set");

	bitvector_set(a, 12);
	bitvector_set(a, 24);
	bitvector_set(a, 48);
	fail_unless(bitvector_firstset(a, &idx) == 0 && idx == 12,
			"Bitvector first set failed when many set");

	bitvector_unset(a, 12);
	fail_unless(bitvector_firstset(a, &idx) == 0 && idx == 24,
			"Bitvector first set failed after unset");

	bitvector_free(a);
}
END_TEST

START_TEST(test_firstset_copy)
{
	bitvector *a = bitvector_create(100);
	bitvector *b = bitvector_create(100);
	index_t idx;

	bitvector_set(a, 12);
	bitvector_set(a, 24);
	bitvector_set(a, 48);
	fail_unless(bitvector_firstset(a, &idx) == 0 && idx == 12,
			"Bitvector first set failed when many set");

	bitvector_copy(a, b);

	fail_unless(bitvector_firstset(b, &idx) == 0 && idx == 12,
			"Bitvector first set failed when many set");

	bitvector_free(a);
	bitvector_free(b);
}
END_TEST

START_TEST(test_firstunset)
{
	bitvector *a = bitvector_create(100);
	index_t idx;

	fail_unless(bitvector_firstunset(a, &idx) == 0 && idx == 0,
			"Bitvector first unset failed when none set");

	bitvector_set(a, 12);
	bitvector_set(a, 24);
	bitvector_set(a, 48);
	fail_unless(bitvector_firstunset(a, &idx) == 0 && idx == 0,
			"Bitvector first unset failed when many set");

	bitvector_set(a, 0);
	bitvector_set(a, 1);
	bitvector_set(a, 2);
	fail_unless(bitvector_firstunset(a, &idx) == 0 && idx == 3,
			"Bitvector first unset failed when many initial set");
}
END_TEST

Suite *bitvector_suite ()
{
	Suite *s = suite_create ("BitVector");
	TCase *tc_core = tcase_create ("Core");

	suite_add_tcase (s, tc_core);

	tcase_add_test (tc_core, test_create);
	tcase_add_test (tc_core, test_create_dirty);
	tcase_add_test (tc_core, test_set);
	tcase_add_test (tc_core, test_unset);
	tcase_add_test (tc_core, test_get);
	tcase_add_test (tc_core, test_clear);
	tcase_add_test (tc_core, test_tostring);
	tcase_add_test (tc_core, test_fromstring);
	tcase_add_test (tc_core, test_fromint);
	tcase_add_test (tc_core, test_resize);
	tcase_add_test (tc_core, test_resize_ns);
	tcase_add_test (tc_core, test_copy);
	tcase_add_test (tc_core, test_duplicate);
	tcase_add_test (tc_core, test_isempty);
	tcase_add_test (tc_core, test_isfull);
	tcase_add_test (tc_core, test_isequal);
	tcase_add_test (tc_core, test_and);
	tcase_add_test (tc_core, test_andeq);
	tcase_add_test (tc_core, test_not);
	tcase_add_test (tc_core, test_invert);
	tcase_add_test (tc_core, test_or);
	tcase_add_test (tc_core, test_oreq);
	tcase_add_test (tc_core, test_xor);
	tcase_add_test (tc_core, test_xoreq);
	tcase_add_test (tc_core, test_leftshift);
	tcase_add_test (tc_core, test_rightshift);
	tcase_add_test (tc_core, test_cstring);
	tcase_add_test (tc_core, test_firstset);
	tcase_add_test (tc_core, test_firstset_copy);
	tcase_add_test (tc_core, test_firstunset);

	return s;
}
#endif

int main (int argc, char **argv)
{
#if HAVE_LIBCHECK
	int nf;
	Suite *s = bitvector_suite ();
	SRunner *sr = srunner_create (s);
	srunner_run_all (sr, CK_NORMAL);
	nf = srunner_ntests_failed (sr);
	srunner_free (sr);
	suite_free (s);

	return (nf == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
#else
	printf ("Not checking, check not present!\n");
	return 0;
#endif
}
